"""
A collection of functions that need to be called via a subprocess call.
"""
from __future__ import annotations

import contextlib
import functools
import importlib.resources
import json
import os
import subprocess
from typing import Any, Generator


@contextlib.contextmanager
def _in_process_script(name: str) -> Generator[str, None, None]:
    with importlib.resources.path(__name__, name) as script:
        yield str(script)


@functools.lru_cache()
def get_python_abi_tag(executable: str) -> str:
    with _in_process_script("get_abi_tag.py") as script:
        return json.loads(subprocess.check_output(args=[executable, "-Es", script]))


def get_sys_config_paths(
    executable: str, vars: dict[str, str] | None = None, kind: str = "default"
) -> dict[str, str]:
    """Return the sys_config.get_paths() result for the python interpreter"""
    env = os.environ.copy()
    env.pop("__PYVENV_LAUNCHER__", None)
    if vars is not None:
        env["_SYSCONFIG_VARS"] = json.dumps(vars)

    with _in_process_script("sysconfig_get_paths.py") as script:
        cmd = [executable, "-Es", script, kind]
        return json.loads(subprocess.check_output(cmd, env=env))


def get_pep508_environment(executable: str) -> dict[str, str]:
    """Get PEP 508 environment markers dict."""
    with _in_process_script("pep508.py") as script:
        args = [executable, "-Es", script]
        return json.loads(subprocess.check_output(args))


def parse_setup_py(executable: str, path: str) -> dict[str, Any]:
    """Parse setup.py and return the kwargs"""
    with _in_process_script("parse_setup.py") as script:
        cmd = [executable, "-Es", script, path]
        return json.loads(subprocess.check_output(cmd))



from __future__ import annotations

from pathlib import Path
from typing import Any, Mapping

import tomlkit
from tomlkit.toml_document import TOMLDocument
from tomlkit.toml_file import TOMLFile

from pdm import termui


class TOMLBase(TOMLFile):
    def __init__(self, path: str | Path, *, ui: termui.UI) -> None:
        super().__init__(path)
        self._path = Path(path)
        self.ui = ui
        self._data = self.read()

    def read(self) -> TOMLDocument:
        if not self._path.exists():
            return tomlkit.document()
        return super().read()

    def set_data(self, data: Mapping[str, Any]) -> None:
        """Set the data of the TOML file."""
        self._data = tomlkit.document()
        self._data.update(data)

    def reload(self) -> None:
        self._data = self.read()

    def write(self) -> None:
        return super().write(self._data)

    @property
    def exists(self) -> bool:
        return self._path.exists()

    @property
    def empty(self) -> bool:
        return not self._data



from pdm.project.toml_file import TOMLBase


class Lockfile(TOMLBase):
    spec_version = "4.1"

    @property
    def hash(self) -> str:
        return self._data.get("metadata", {}).get("content_hash", "")

    @property
    def file_version(self) -> str:
        return self._data.get("metadata", {}).get("lock_version", "")

    def write(self, show_message: bool = True) -> None:
        super().write()
        if show_message:
            self.ui.echo(f"Changes are written to [success]{self._path.name}[/].")

    def __getitem__(self, key: str) -> dict:
        return self._data[key]



import importlib.resources

from pdm.compat import importlib_metadata


def read_version() -> str:
    try:
        return importlib_metadata.version(__package__ or "pdm")
    except importlib_metadata.PackageNotFoundError:
        return importlib.resources.read_text("pdm.models", "VERSION").strip()


__version__ = read_version()



from __future__ import annotations

from unearth.auth import MultiDomainBasicAuth

from pdm._types import Source
from pdm.exceptions import PdmException
from pdm.termui import UI

try:
    import keyring
except ModuleNotFoundError:
    keyring = None  # type: ignore

ui = UI()


class PdmBasicAuth(MultiDomainBasicAuth):
    """A custom auth class that differs from Pip's implementation in the
    following ways:

        - It shows an error message when credentials are not provided or correct.
    """

    def __init__(
        self, prompting: bool = True, index_urls: list[str] | None = None
    ) -> None:
        super().__init__(prompting=True, index_urls=index_urls)
        self._real_prompting = prompting

    def _prompt_for_password(self, netloc: str) -> tuple[str | None, str | None, bool]:
        if not self._real_prompting:
            raise PdmException(
                f"The credentials for {netloc} are not provided. "
                "Please rerun the command with `-v` option."
            )
        return super()._prompt_for_password(netloc)

    def _should_save_password_to_keyring(self) -> bool:
        if keyring is None:
            ui.echo(
                "The provided credentials will not be saved into your system.\n"
                "You can enable this by installing keyring:\n"
                "    pdm self add keyring",
                err=True,
                style="warning",
            )
        return super()._should_save_password_to_keyring()


def make_basic_auth(sources: list[Source], prompting: bool) -> PdmBasicAuth:
    return PdmBasicAuth(
        prompting,
        [source["url"] for source in sources if source.get("type", "index") == "index"],
    )



import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


if sys.version_info >= (3, 8):
    from functools import cached_property
    from typing import Literal, Protocol, TypedDict
else:
    from typing import Any, Callable, Generic, TypeVar, overload

    from typing_extensions import Literal, Protocol, TypedDict

    _T = TypeVar("_T")
    _C = TypeVar("_C")

    class cached_property(Generic[_T]):
        def __init__(self, func: Callable[[Any], _T]):
            self.func = func
            self.attr_name = func.__name__
            self.__doc__ = func.__doc__

        @overload
        def __get__(self: _C, inst: None, cls: Any = ...) -> _C:
            ...

        @overload
        def __get__(self, inst: object, cls: Any = ...) -> _T:
            ...

        def __get__(self, inst, cls=None):
            if inst is None:
                return self
            if self.attr_name not in inst.__dict__:
                inst.__dict__[self.attr_name] = self.func(inst)
            return inst.__dict__[self.attr_name]


if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata


Distribution = importlib_metadata.Distribution


__all__ = [
    "tomllib",
    "cached_property",
    "importlib_metadata",
    "Literal",
    "Protocol",
    "TypedDict",
    "Distribution",
]



from __future__ import annotations

import hashlib
import json
from typing import Mapping

from tomlkit import TOMLDocument, items

from pdm.project.toml_file import TOMLBase
from pdm.utils import deprecation_warning


class PyProject(TOMLBase):
    """The data object representing th pyproject.toml file"""

    def read(self) -> TOMLDocument:
        from pdm.formats import flit, poetry

        data = super().read()
        if "project" not in data and self._path.exists():
            # Try converting from flit and poetry
            for converter in (flit, poetry):
                if converter.check_fingerprint(None, self._path):
                    metadata, settings = converter.convert(None, self._path, None)
                    data["project"] = metadata
                    if settings:
                        data.setdefault("tool", {}).setdefault("pdm", {}).update(
                            settings
                        )
                    break
        return data

    def write(self, show_message: bool = True) -> None:
        """Write the TOMLDocument to the file."""
        super().write()
        if show_message:
            self.ui.echo("Changes are written to [success]pyproject.toml[/].")

    @property
    def is_valid(self) -> bool:
        return "project" in self._data

    @property
    def metadata(self) -> items.Table:
        return self._data.setdefault("project", {})

    @property
    def settings(self) -> items.Table:
        return self._data.setdefault("tool", {}).setdefault("pdm", {})

    @property
    def build_system(self) -> dict:
        return self._data.get("build-system", {})

    @property
    def resolution_overrides(self) -> Mapping[str, str]:
        """A compatible getter method for the resolution overrides
        in the pyproject.toml file.
        """
        settings = self.settings
        if "overrides" in settings:
            deprecation_warning(
                "The 'tool.pdm.overrides' table has been renamed to "
                "'tool.pdm.resolution.overrides', please update the "
                "setting accordingly."
            )
            return settings["overrides"]
        return settings.get("resolution", {}).get("overrides", {})

    def content_hash(self, algo: str = "sha256") -> str:
        """Generate a hash of the sensible content of the pyproject.toml file.
        When the hash changes, it means the project needs to be relocked.
        """
        dump_data = {
            "sources": self.settings.get("source", []),
            "dependencies": self.metadata.get("dependencies", []),
            "dev-dependencies": self.settings.get("dev-dependencies", {}),
            "optional-dependencies": self.metadata.get("optional-dependencies", {}),
            "requires-python": self.metadata.get("requires-python", ""),
            "overrides": self.resolution_overrides,
        }
        pyproject_content = json.dumps(dump_data, sort_keys=True)
        hasher = hashlib.new(algo)
        hasher.update(pyproject_content.encode("utf-8"))
        return hasher.hexdigest()



import argparse
import sys

from pdm import termui
from pdm.cli import actions
from pdm.cli.commands.base import BaseCommand
from pdm.cli.hooks import HookManager
from pdm.cli.options import (
    dry_run_option,
    groups_group,
    install_group,
    lockfile_option,
    skip_option,
)
from pdm.project import Project


class Command(BaseCommand):
    """Install dependencies from lock file"""

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        groups_group.add_to_parser(parser)
        install_group.add_to_parser(parser)
        dry_run_option.add_to_parser(parser)
        lockfile_option.add_to_parser(parser)
        skip_option.add_to_parser(parser)
        parser.add_argument(
            "--no-lock",
            dest="lock",
            action="store_false",
            default=True,
            help="Don't do lock if the lock file is not found or outdated",
        )
        parser.add_argument(
            "--check",
            action="store_true",
            help="Check if the lock file is up to date and fail otherwise",
        )

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        if not project.pyproject.is_valid and termui.is_interactive():
            actions.ask_for_import(project)

        hooks = HookManager(project, options.skip)

        strategy = actions.check_lockfile(project, False)
        if strategy:
            if options.check:
                project.core.ui.echo(
                    "Please run [success]`pdm lock`[/] to update the lock file",
                    err=True,
                )
                sys.exit(1)
            if options.lock:
                project.core.ui.echo(
                    "Updating the lock file...", style="success", err=True
                )
                actions.do_lock(
                    project, strategy=strategy, dry_run=options.dry_run, hooks=hooks
                )

        actions.do_sync(
            project,
            groups=options.groups,
            dev=options.dev,
            default=options.default,
            no_editable=options.no_editable,
            no_self=options.no_self,
            dry_run=options.dry_run,
            hooks=hooks,
        )
